import { EventEmitter } from '../Events/EventEmitter.js';
import { NodeEvaluationEvent } from '../Graphs/Evaluation/NodeEvaluationEvent.js';
import { NodeEvaluationType } from '../Graphs/Evaluation/NodeEvaluationType.js';
import { Link } from './Link.js';
export class NodeEvalContext {
    constructor(syncExecutionBlock, node) {
        this.syncExecutionBlock = syncExecutionBlock;
        this.node = node;
        this.onAsyncCancelled = new EventEmitter();
        this.cachedInputValues = {};
        this.cachedOutputValues = {};
        this.asyncPending = false;
        this.numCommits = 0;
        this.graphEvaluator = syncExecutionBlock.graphEvaluator;
        this.graph = this.graphEvaluator.graph;
    }
    begin() {
        if (this.node.interruptibleAsync) {
            this.graphEvaluator.interruptibleAsyncNodes.push(this.node);
        }
        else {
            this.graphEvaluator.asyncNodes.push(this.node);
        }
        this.asyncPending = true;
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.Flow, true));
    }
    cancel() {
        this.onAsyncCancelled.emit();
        this.finish();
    }
    finish() {
        if (this.node.interruptibleAsync) {
            const index = this.graphEvaluator.interruptibleAsyncNodes.indexOf(this.node);
            this.graphEvaluator.interruptibleAsyncNodes.splice(index, 1);
        }
        else {
            const index = this.graphEvaluator.asyncNodes.indexOf(this.node);
            this.graphEvaluator.asyncNodes.splice(index, 1);
        }
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.None, false));
    }
    evalFlow() {
        if (this.node.async) {
            this.begin();
        }
        else {
            this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.Flow, false));
        }
        this.readInputs();
        this.node.evalFunc(this);
        if (!this.node.async) {
            this.writeOutputs();
            this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.None, false));
        }
    }
    evalImmediate() {
        if (this.node.flow) {
            throw new Error('can not evalImmediate on Flow nodes, use evalFlow instead');
        }
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.Immediate, false));
        this.readInputs();
        this.node.evalFunc(this);
        this.writeOutputs();
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(this.node, NodeEvaluationType.None, false));
    }
    readInputs() {
        this.node.inputSockets.forEach((socket) => {
            this.cachedInputValues[socket.name] = socket.value;
        });
    }
    writeOutputs() {
        this.node.outputSockets.forEach((socket) => {
            socket.value = this.cachedOutputValues[socket.name];
        });
    }
    getCustomEvent(customEventId) {
        if (!(customEventId in this.graph.customEvents)) {
            throw new Error(`can not find customEvent with the id ${customEventId}`);
        }
        return this.graph.customEvents[customEventId];
    }
    getVariable(variableId) {
        if (!(variableId in this.graph.variables)) {
            throw new Error(`can not find variable with the id ${variableId}`);
        }
        return this.graph.variables[variableId];
    }
    readInput(inputName) {
        const inputSocket = this.node.inputSockets.find((socket) => socket.name === inputName);
        if (inputSocket === undefined) {
            throw new Error(`can not find input socket with name ${inputName}`);
        }
        return this.cachedInputValues[inputName];
    }
    writeOutput(outputName, value) {
        const outputSocket = this.node.outputSockets.find((socket) => socket.name === outputName);
        if (outputSocket === undefined) {
            throw new Error(`can not find output socket with name ${outputName} on node of type ${this.node.description.typeName}`);
        }
        if (outputSocket.valueTypeName === 'flow') {
            throw new Error(`can not set the value of Flow output socket ${outputName}, use commit() instead`);
        }
        this.cachedOutputValues[outputName] = value;
    }
    commit(downstreamFlowSocketName, syncEvaluationCompletedListener = undefined) {
        this.numCommits++;
        this.writeOutputs();
        if (this.node.async) {
            this.graphEvaluator.asyncCommit(new Link(this.node.id, downstreamFlowSocketName), syncEvaluationCompletedListener);
        }
        else {
            this.syncExecutionBlock.commit(new Link(this.node.id, downstreamFlowSocketName), syncEvaluationCompletedListener);
        }
    }
}
//# sourceMappingURL=NodeEvalContext.js.map